/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

class CodeSpace {
	private class ProgInfo {
		private int id;
		private int maxTasks;
		private FragInfo fragments[];
		private int start;
		private int end;

		ProgInfo(int id, int maxTasks, int maxSubs) {
			this.id = id;
			this.maxTasks = maxTasks;
			
			int n = maxTasks + maxSubs;
			fragments = new FragInfo[n];
			
			for(int i=0; i<n; ++i)
				fragments[i] = new FragInfo();
		}
		
		int getId() { return id; }
		int getStart() { return start; }
		int getEnd() { return end; }
		
		FragInfo getFragInfo(int type, int index) {
			if (type == Fragment.SUB_FRAGMENT)
				index += maxTasks;
			return fragments[index];
		}
		
		void setEnd(int e) { end = e; }
		
		void clear() {
			end = start;
			for(int i=0; i<fragments.length; ++i)
				fragments[i].clear(start);
		}
		
		void move(int delta) {
			start += delta;
			end += delta;
			for(int i=0; i<fragments.length; ++i)
				fragments[i].move(delta);
		}
	}


	private class FragInfo extends Range {
		void setLocation(int s, int e) {
			start = s;
			end = e;
		}
		
		void clear(int s) {
			start = s;
			end = s;
		}
		
		void move(int delta) {
			start += delta;
			end += delta;
		}
	}


	private byte code[];
	private ProgInfo programs[];
	
	CodeSpace(int size, int programCount, int maxTasks, int maxSubs) {
		code = new byte[size];
		programs = new ProgInfo[programCount];
		for(int i=0; i<programCount; ++i)
			programs[i] = new ProgInfo(i, maxTasks, maxSubs);
	}
	
	byte[] getCode() { return code; }
	

	void load(int id, Program program) {
		clear(id);
		
		// compute total size of new fragments;
		int newSize = 0;
		for(int i=0; i<program.getFragmentCount(); ++i)
			 newSize += program.getFragment(i).getCode().length;

		// adjust other programs
		for(int i=programs.length-1; i>id; --i)
			move(programs[i], newSize);
		
		ProgInfo pi = programs[id];
		int pos = pi.getStart();
		for(int i=0; i<program.getFragmentCount(); ++i) {
			Fragment f = program.getFragment(i);
			byte src[] = f.getCode();
			int length = src.length;
			
			for(int j=0; j<length; ++j)
				code[pos + j] = src[j];
			
			FragInfo fi = pi.getFragInfo(f.getType(), f.getId());
			
			fi.setLocation(pos, pos+length);
			pos += length;
		}
		pi.setEnd(pos);
	}

	Range getFragmentRange(int program, int type, int index) {
		return programs[program].getFragInfo(type, index);
	}

		
	private void clear(int id) {
		int start = programs[id].getStart();
		int end = programs[id].getEnd();
		int adjust = start - end;
		
		// no space...just return
		if (adjust==0) return;
		
		programs[id].clear();
		
		for(int i=id+1; i<programs.length; ++i)
			move(programs[i], adjust);
	}

	private void move(ProgInfo p, int delta) {
		if (delta == 0) return;
		
		if (delta < 0) {
			for(int i=p.getStart(); i<p.getEnd(); ++i)
				code[i+delta] = code[i];
		}
		else {
			for(int i=p.getEnd()-1; i>=p.getStart(); --i)
				code[i+delta] = code[i];
		}
		
		p.move(delta);
	}
}
