/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

public class Display {
	// the following comments must match those in the RCX itself
	public static final int DISPLAY_WATCH = 0;
	public static final int DISPLAY_SENSOR_1 = 1;
	public static final int DISPLAY_SENSOR_2 = 2;
	public static final int DISPLAY_SENSOR_3 = 3;
	public static final int DISPLAY_OUT_A = 4;
	public static final int DISPLAY_OUT_B = 5;
	public static final int DISPLAY_OUT_C = 6;
	public static final int DISPLAY_USER = 7;
	
	// number of characters in the display
	private static final int DIGITS = 4;
	public static final int WIDTH = DIGITS + 1; // room for leading '-'

	
	private DisplayListener listener;
	private int				mode = DISPLAY_WATCH;
	private boolean			formatChanged;
	private int				actualValue;
	private int				dValue;
	private int				point = 2;
	private boolean			zeroFill = true;
	private char			chars[];
	private int				userSource;
	private int				userIndex;
	private int				userPoint;
	
	private PBrick			pb;

	Display(PBrick pb) {
		this.pb = pb;
		chars = new char[WIDTH];
		setValue(0);
	}

	public int getMode() { return mode; }
	public int getValue() { return dValue; }
	public int getPoint() { return point; }
	public char getChar(int i) 	{ return chars[i]; }

	public String toString() {
		StringBuffer sb = new StringBuffer();
		
		int leading = WIDTH - point;
		for(int i=0; i<leading; ++i)
			sb.append(chars[i]);
		
		if (point > 0)
			sb.append('.');
		
		for(int i=leading; i<WIDTH; ++i)
			sb.append(chars[i]);
			
		return sb.toString();
	}
	
	public void setListener(DisplayListener l) { listener = l; }
	
	
	void setUserDisplay(int point, int src, int index) {
		userSource = src;
		userIndex = index;
		userPoint = point;
		mode = DISPLAY_USER;
		useUserPoint();
		formatChanged = true;

		if (listener != null) {
			listener.pbDisplayModeChanged(this);
		}
		
		refresh();
	}
	
	void setMode(int m) {
		if (mode == m) return;
		
		mode = m;
		switch(mode) {
			case DISPLAY_WATCH:
			default:
				point = 2;
				zeroFill = true;
				break;
			case DISPLAY_SENSOR_1:
			case DISPLAY_SENSOR_2:
			case DISPLAY_SENSOR_3:
			case DISPLAY_OUT_A:
			case DISPLAY_OUT_B:
			case DISPLAY_OUT_C:
				point = 0;
				zeroFill = false;
				break;
			case DISPLAY_USER:
				useUserPoint();
				break;
		}

		if (listener != null) {
			listener.pbDisplayModeChanged(this);
		}
		
		formatChanged = true;
		refresh();
	}


	private void useUserPoint() {
		if (userPoint > 3 || userPoint < 0) {
			point = 0;
			zeroFill = true;
		}
		else {
			point = userPoint;
			zeroFill = false;
		}
	}	
	
	
	void refresh() {
		int v;
		int newPoint;
		int sensorMode;
		Sensor sensor;
		
		switch(mode) {
			case DISPLAY_WATCH:
			default:
				v = pb.getWatch();
				v += (v / 60) * 40;
				break;
			case DISPLAY_SENSOR_1:
			case DISPLAY_SENSOR_2:
			case DISPLAY_SENSOR_3:
				sensor = pb.getSensor(mode-DISPLAY_SENSOR_1);
				v = sensor.getValue();
				sensorMode = sensor.getMode();
				newPoint = (sensorMode == Sensor.MODE_CELSIUS || sensorMode == Sensor.MODE_FAHRENHEIT) ? 1 : 0;
				if (point != newPoint) {
					point = newPoint;
					formatChanged = true;
				}
				break;
			case DISPLAY_OUT_A:
			case DISPLAY_OUT_B:
			case DISPLAY_OUT_C:
				v = pb.getOutput(mode-DISPLAY_OUT_A).getViewValue();
				break;
			case DISPLAY_USER:
				v = pb.evaluate(userSource, userIndex);
				break;
		}
		
		if (v != actualValue || formatChanged) {
			setValue(v);
			if (listener != null)
				listener.pbDisplayValueChanged(this);
		}
		
		formatChanged = false;
	}


	private void setValue(int v) {
		actualValue = v;

		if (v < -9999)
			v = -9999;
		else if (v > 9999)
			v = 9999;
			
		dValue = v;
		
		// format as a positive value
		if (v < 0)
			v = -v;
		
		chars[0] = ' ';
		for(int i=0; i<DIGITS; ++i) {
			chars[WIDTH-1-i] = (char)('0' + (v % 10));
			v /= 10;
		}
		
		// trim zeros?
		int first = 0;
		if (!zeroFill) {
			for(first=0; first< DIGITS-point-1; ++first)
				if (chars[first+1] != '0')
					break;
				else
					chars[first+1] = ' ';
		}
		if (dValue < 0)
			chars[first] = '-';
	}

}
