/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

public class Output
{
	// output directions - MUST MATCH RCX BYTECODE VALUES
	public static final int OUT_REV = 0;
	public static final int OUT_TOGGLE = 1;
	public static final int OUT_FWD = 2;
	
	// output modes - MUST MATCH RCX BYTECODE VALUES
	public static final int OUT_FLOAT = 0;
	public static final int OUT_OFF = 1;
	public static final int OUT_ON = 2;

	// power
	public static final int OUT_FULL = 7;
	
	private int id;
	
	// actual state of output
	private int mode = OUT_OFF;
	private int power = OUT_FULL;
	private boolean forward = true;
	
	// global state
	private int maxPower = OUT_FULL;
	private boolean disabled = false;	// motor mode cannot be changed
	private boolean reversed = false;	// motor directions swappped
	
	private OutputListener listener;

	// public accessors
	public int		getId()			{ return id; }
	public int		getMode()		{ return mode; }
	public boolean	getForward()	{ return forward; }
	public int 		getPower()		{ return power; }

	// package can change an output
	Output(int id) {
		this.id = id;
		listener = null;
	}
				
	public void	setListener(OutputListener l) { listener = l; }

	void	reset() {
		mode = OUT_OFF;
		forward = true;
		power = OUT_FULL;
		maxPower = OUT_FULL;
		disabled = false;
		reversed = false;
		if (listener != null) listener.pbOutputChanged(this);
	}
	
	void	setMode(int m) {
		if (disabled) return;
		if (mode == m) return;
		mode = m;
		changed();
	}
	
	void	setDirection(int d)	{
		boolean f;
		
		switch(d) {
			case OUT_FWD:
				f = !reversed;
				break;
			case OUT_REV:
				f = reversed;
				break;
			case OUT_TOGGLE:
				f = !forward;
				break;
			default:
				return;
		}
				
		if (f == forward) return;
		
		forward = f;
		if (listener != null) listener.pbOutputChanged(this);
	}
	
	void	setPower(int p) {
		if (p > maxPower) p = maxPower;
		if (power == p) return;
		power = p;
		changed();
	}
	
	
	void setGlobalMode(int m) {
		if (m == OUT_ON) {
			disabled = false;
		}
		else {
			disabled = true;
			if (mode != m) {
				mode = m;
				changed();
			}
		}
	}
	
	
	void setGlobalDirection(int d) {
		boolean r;
		
		switch(d) {
			case OUT_FWD:
				r = false;
				break;
			case OUT_REV:
				r = true;
				break;
			case OUT_TOGGLE:
				r = !reversed;
				break;
			default:
				return;
		}
		
		if (reversed == r) return;
		reversed = r;
		forward = !forward;
		changed();
	}
	
	
	void setGlobalPower(int p) {
		maxPower = p;
		if (power > maxPower) {
			power = maxPower;
			changed();
		}
	}
	
	private void changed() {
		if (listener != null) listener.pbOutputChanged(this);
	}
	
	
	int		getViewValue() {
		if (mode == OUT_ON)
			return power + 1;
		else
			return 0;
	}
	
	int		getStatus() {
		int status = (mode << 6)  + power;
		if (forward) status += 0x08;
		return status;
	}
	
	int		getGlobalStatus() {
		int status = power + 0x30;
		if (!reversed) status += 0x08;
		if (!disabled) status += 0x80;
		return status;
	}
};

