/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

/** This class provides the external API to the core simulation.
 * <p>
 *	Clients can not instantiate this class directly - they should instantiate
 *  one of the derrived classes (e.g. RCX) that implements the feature set
 *  of a specific programmable brick.
 *	<p>
 *	PBrick relies heavily on a Listener model to inform a client of various
 *	events during the simulation.  For example, if the client wants to know
 *	when an output changes state, it should implement the OutputListener
 *	interface and call setOutputListener().
 *	It is not necessary for a client to implement all of the listeners, nor
 *	do they have to all be the same object.
 *	<p>
 *	The simulation maintains "simulation time", which is potentially much
 *	different from real time.  For example, my machine runs the simulation
 *	about 350 times as fast as a real RCX would, so one second of real time
 *	is about 6 minutes of simulation time.  In some cases it may be desireable
 *	to make the simulation approximate real time.  This can be done by
 *	interspersing calls to Thread.sleep() with calls to PBrick.step().
 */

import java.util.Random;

public class PBrick
{
	public static final int MAX_INT = 32767;
	public static final int MIN_INT = -32768;
	
	private static final int MICROS_PER_INSTRUCTION = 2000; // 2 ms
	private static final int MICROS_PER_TICK = 10000; // 10ms
	private static final int TICKS_PER_SECOND = 100;
	
	private static final int FIRMWARE_VERSION = 328;
	private static final int BATTERY_LEVEL = 9000;
	
	static final int MAX_RESOURCES = 8;
	
	// opcode map
	private boolean		opcodeLegal[] = new boolean[256];

	private static int coreOpcodes[] = {
		Op.START, Op.STOP, Op.STOPA, Op.WAIT,
		Op.SETV, Op.SUMV, Op.SUBV, Op.DIVV, Op.MULV, Op.SGNV, Op.ABSV, Op.ANDV, Op.ORV,
		Op.OUT, Op.DIR, Op.PWR, 
		Op.JUMP, Op.JUMPL, Op.CHK, Op.CHKL, Op.CALLS, Op.RTS,
		Op.PLAYS, Op.PLAYT, /* Op.OFFP, */
		Op.TMRZ, Op.TOUT,
	};
	
	// outputs	
	private	Output			outputs[];

	// variables
	private Storage		storage;
	private DebugListener	debugListener;

	// programs
	private CodeSpace	codeSpace;
	private int			currentProgram;

	// messages
	private MessageListener messageListener;
	private int				message;
	private TriggerQueue	messageTriggers;

	// sensors
	private boolean		sampleSensors;	
	private Sensor		sensors[];

	// other
	private SoundListener	soundListener;
	private Counter		counters[];
	private Task		currentTask;
	private int			currentPc;
	private Datalog		datalog;
	private Random		random;

	private Scheduler	scheduler;
	
	// loop levels
	private int			loopCounters[];
	private int			loopDepth;
		
	// timing
	private Timer		timers[];
	private int			microCounter;
	private int			ticks;
	private int			watch;
	private int			watchPhase;
	
	// events
	private Trigger		triggers[];
	private int			pendingEvents;
	
	// display
	private	Display		display;
	
	
	PBrick(int codeSize, int programCount, int taskCount, int subCount, int globalCount, int localCount) {
		random = new Random();
		storage = new Storage(globalCount, localCount, taskCount);
		
		codeSpace = new CodeSpace(codeSize, programCount, taskCount, subCount);
		
		scheduler = new Scheduler(this, taskCount, codeSpace.getCode());
		
		messageTriggers = new TriggerQueue();
		enableOpcodes(coreOpcodes);
	}
	
	
	Scheduler getScheduler() { return scheduler; }

	
	/** Select which program slot should be "active".  This has the side
	 *	effect of stopping the currently running program (if any).
	 *	@param n desired program slot (0 based)
	 */
	public void switchProgram(int n) {
		stop();
		currentProgram = n;
	}

	/** Load a program into the "active" slot.  This has the side
	 *	effect of stopping the currently running program (if any).  Data
	 *	is read and copied out of p - any future modification to p will not
	 *	affect the running of the program within the PBrick.
	 *	@param p new program to load
	 */
	public void load(Program p) {
		stop();
		codeSpace.load(currentProgram, p);
	}


	/** Set the listener to be used for Sound events */
	public void setSoundListener(SoundListener s) { soundListener = s; }
	
	/** Set the listener to be used for Debug events */
	public void setDebugListener(DebugListener l) {
		debugListener = l;
		storage.setDebugListener(l);
		scheduler.setDebugListener(l);
	}


	protected void createDatalog() {
		datalog = new Datalog(this);
		enableOpcode(Op.LOG);
		enableOpcode(Op.LOGZ);
		enableOpcode(Op.POLLD);
	}
	
	/** Get the datalog.
	 *	@return the datalog, or null if PBrick doesn't support a datalog
	 */
	public Datalog getDatalog() { return datalog; }


	protected void createCounters(int n, boolean overlap) {
		counters = new Counter[n];
		for(int i=0; i<n; ++i)
			counters[i] = new Counter();

		enableOpcode(Op.CNTD);
		enableOpcode(Op.CNTI);
		enableOpcode(Op.CNTZ);
		
		if (overlap)
			storage.overlapCounters(counters);
	}


	void releaseResource(int r) {
		switch(r) {
			case 0:
				outputs[0].setMode(Output.OUT_OFF);
				break;
			case 1:
				outputs[1].setMode(Output.OUT_OFF);
				break;
			case 2:
				if (soundListener != null) soundListener.pbControlSound(SoundListener.CLEAR);					
				break;
			case 3:
				outputs[2].setMode(Output.OUT_OFF);
				break;
		}
	}


	/*
	 * Communication
	 */

	/** Set the listener to be used for Message events */
	public synchronized void setMessageListener(MessageListener m) { messageListener = m; }


	/** Deliver a message to the PBrick.  Use this call to simulate a host computer
	 *	or another PBrick sending a message to this one.
	 *	@param m value of the message (0-255)
	 */
	public synchronized void deliverMessage(int m) {
		// clip m
		m = m & 0xff;
		message = m;
		messageTriggers.update(m, 1 << Trigger.EVENT_TYPE_MESSAGE);
	}


	void addMessageTrigger(Trigger t) {
		t.enqueue(messageTriggers);
		t.prime(message, 0, 255);
	}


	void sendMessage(int m) {
		if (messageListener != null)
			messageListener.pbMessageSent(m & 0xff);
	}

	/*
	 * Tasks/Timing
	 */
	
	/** Start the current program.  If the program is already runniing, it will
	 *	be stopped, then restarted.  This is similar to pressing the "Run" button
	 *	on a real RCX when a program is not already running.  Note that this will
	 *	not cause the simulator to actually execute any program code - it just
	 *	puts the simulator in a mode where the tasks are ready to be run.  The
	 *	methods step() and runFor() are used to advance the simulation.
	 */
	public synchronized void start() {
		stop();
		start(0);
	}


	/** Stop the current program.  This stops all tasks in the program, putting
	 *  the PBrick back in the idle state.  This is similar to pressing the "Run"
	 *	button on a real RCX when a program is running.
	 */
	public synchronized void stop() {
		scheduler.stopAll();
		for(int i=0; i<outputs.length; ++i)
			outputs[i].setMode(Output.OUT_OFF);
	}
	
	
	/** Check if a program is running.
	 *	@returns true iff a program is running
	 */
	public boolean isRunning() { return scheduler.isRunning(); }


	/** Get the system tick count.
	 *	The PBrick uses a system tick of 10ms (simulation time, not real time).  This
	 *	method returns the current system tick count.  It can be used to determine
	 *	how much simulation time has passed.
	 */
	public int	getTicks() { return ticks; }


	/** Get the system watch.  The system watch keeps track of the number of minutes
	*	since reset.  It wraps after 24 hours.
	*	@returns the system watch value
	*/
	public int getWatch() { return watch; }
	
	
	/** Reset the PBrick.
	 *	Stops the current program and resets all timing values.  This is similar
	 * 	to turning a real RCX off and back on.
	 */
	 public void reset() {
	 	stop();
	 	
	 	microCounter = 0;
	 	ticks = 0;
	 	watch = 0;
	 	watchPhase = 0;
	 	for(int i=0; i<timers.length; ++i)
	 		timers[i].clear();
	 	
	 	for(int i=0; i<outputs.length; ++i) {
	 		outputs[i].reset();
	 	}
	 	
	 	pendingEvents = 0;
	 }
	
	/** Execute one bytecode instruction.
	 *	One instruction is simulated, possibly resulting in callbacks
	 *	to listeners and/or advancing the system time.  This is the primary
	 *	mechanism for advancing the simulation.
	 */

	public synchronized void step() {
		if (sampleSensors) {
			sampleSensors = false;
			for(int i=0; i<sensors.length; ++i)
				sensors[i].sample();
		}
		
		if (execute()) {
			microCounter += MICROS_PER_INSTRUCTION;
			while(microCounter > 0) {
				tick();
				microCounter -= MICROS_PER_TICK;
			}
		}
		else {
			tick();
		}
	}
	
	/** Execute for a specified number of simulation ticks.
	 *	The simulation is run until the specified number of ticks has
	 *	elapsed (in simulation time, not real time).
	 *	@param duration amount of time in simulated 10ms ticks
	 */
	public synchronized void runFor(int duration) {
		int target = ticks + duration;
		while(ticks < target)
			step(); 
	}


	protected void createTimers(int n) {
		timers = new Timer[n];
		for(int i=0; i<n; ++i) {
			timers[i] = new Timer(i);
		}
	}
	
	
	protected Timer getTimer(int i) { return timers[i]; }
	
	private void start(int id) {
		scheduler.start(id, codeSpace.getFragmentRange(
			currentProgram, Fragment.TASK_FRAGMENT,id));
	}
	
		
	private void tick() {
		// this is the 10ms system tick
		
		ticks++;
		scheduler.wakeup(ticks);
		
		// tick the timers
		for(int i=0; i<timers.length; ++i)
			timers[i].tick();
		
		// divide by 100 for the watch tick
		if (++watchPhase == TICKS_PER_SECOND * 60) {
			watchPhase = 0;
			++watch;
			if (watch == 24 * 60) watch = 0;
		}
		
		display.refresh();
	}
	
	
	private void setWatch(int h, int m) {
		if (m < 0 || m > 59) return;
		if (h < 0 || h > 23) return;
		
		watch = h * 60 + m;
		watchPhase = 0;
	}
	
	
	/*
	 * Flow control
	 */
	 
	protected void createLoopCounters(int n) {
		loopCounters = new int[n];
		loopDepth = 0;
		enableOpcode(Op.LOOPC);
		enableOpcode(Op.LOOPCL);
		enableOpcode(Op.LOOPS);
	}
	
	
	private void setLoopCounter(int v) {
		if (loopCounters == null) return;
		if (loopDepth == loopCounters.length) return;
		
		if (v<0 || v>255) v = 0;
		loopCounters[loopDepth++] = v;
	}

	private void execCheckLoop(boolean longJump) throws TaskException {
		boolean jump = false;
		
		if (loopDepth > 0) {
			int index = loopDepth - 1;		
			if (loopCounters[index] == 0) {
				loopDepth--;
				jump = true;
			}
			else
				loopCounters[index]--;
		}

		int target = currentTask.nextSignBitTarget(longJump);
		if (jump)
			currentTask.setPC(target);
	}
	
	
	private void execDecvjn(boolean longJump) throws TaskException {
		Task t = currentTask;
		
		int index = t.nextByte();
		short value = storage.get(index, t.getId()); 
		--value;
		storage.set(index, t.getId(), value);
		
		int target = currentTask.nextSignBitTarget(longJump);
		if (value < 0)
			currentTask.setPC(target);			
	}
	
	
	private void execCheck(boolean longJump) throws TaskException {
		Task t = currentTask;
		int b = t.nextByte();
		int comp = b >> 6;
		int src1 = b & 0x3f;
		int src2 = t.nextByte();
		int val1 = t.nextWord();
		int val2 = t.nextByte();
		
		int v1 = evaluate(src1, val1);
		int v2 = evaluate(src2, val2);
		
		boolean branch = false;  // assign default to keep compiler happy
		
		switch(comp) {
			case 0: // <=
				branch = (v1 <= v2);
				break;	
			case 1: // >=
				branch = (v1 >= v2);
				break;
			case 2: // !=
				branch = (v1 != v2);
				break;
			case 3: // ==
				branch = (v1 == v2);
				break;
		}
		
		int target = t.nextOffsetTarget(longJump);
		if (branch)
			t.setPC(target);
	}
	
	
	/*
	 * Sensors
	 */

	/** Get the number of sensors supported by the PBrick.
	 *	@return number of sensors
	 */
	public int getSensorCount() { return sensors.length; }
	
	/** Get a specified sensor.
	 *	@param i index of the sensor (0 based)
	 *	@return the specified sensor
	 */ 
	public Sensor getSensor(int i) { return sensors[i]; }
	

	/** Set the listener to be used for sensor events */
	public void setSensorListener(SensorListener l) {
		for(int i=0; i<sensors.length; ++i)
			sensors[i].setListener(l);
	}
	
	void setSampleSensors()	{ sampleSensors = true; }
	
	protected void createSensors(int n) {
		sensors = new Sensor[n];
		for(int i=0; i<n; ++i) {
			sensors[i] = new Sensor(this, i);
		}
		
		sampleSensors = false;
	}


	/*
	 * Outputs
	 */
	 
	/** Get the number of outputs supported by the PBrick.
	*	@return number of outputs
	*/
	public int getOutputCount() { return outputs.length; }


	/** Get a specified output.
	*	@param i index of the output (0 based)
	*	@return the specified sensor
	*/
	public Output getOutput(int i)	{ return outputs[i]; }

	/** Set the listener to be used for output events */
	public void setOutputListener(OutputListener l) {
		for(int i=0; i<outputs.length; ++i)
			outputs[i].setListener(l);
	}
	
	protected void createOutputs(int n) {
		outputs = new Output[n];
		for(int i=0; i<n; ++i)
			outputs[i] = new Output(i);
	}

	private void execOutputs(int opcode, int arg1, int arg2) {
		for(int i=0; i<3; ++i) {
			if ((arg1 & (1 << i)) == 0) continue;
			Output out = outputs[i];
			
			switch(opcode) {
				case Op.OUT:
					out.setMode(arg1 >> 6);
					break;
				case Op.DIR:
					out.setDirection(arg1 >> 6);
					break;
				case Op.PWR:
					out.setPower(arg2);
					break;
				case Op.GOUT:
					out.setGlobalMode(arg1 >> 6);
					break;
				case Op.GDIR:
					out.setGlobalDirection(arg1 >> 6);
					break;
				case Op.GPWR:
					out.setGlobalPower(arg2);
					break;
			}
		}
	}
	
	
	/*
	 * Display
	 */
	
	/** Set the listener used for display events. */
	public void setDisplayListener(DisplayListener l) { display.setListener(l); }
	
	public Display getDisplay() { return display; }
	
	protected void createDisplay() {
		display = new Display(this);
		enableOpcode(Op.VIEW);
	}
	

	/*
	 * Events
	 */
	
	protected void createEvents(int count) {
		triggers = new Trigger[count];
		for(int i=0; i<count; ++i)
			triggers[i] = new Trigger(this, i);
	}
	
	protected Trigger getTrigger(int i) { return triggers[i]; }
	
	synchronized void triggerEvents(int mask) {
		pendingEvents |= mask;
	}


	private void clearAllEvents() {
		for(int i=0; i<triggers.length; ++i)
			triggers[i].enqueue(null);
	}


	private void setEvent(int evt, int evtSrc, int evtType) {
		Trigger t = triggers[evt];

		if (evtType == 16) {
			// clear the event
			t.enqueue(null);
			return;
		}
	
		t.setType(evtType);
		
		switch(evtSrc) {
			case 0:
			case 1:
			case 2:
				sensors[evtSrc].addTrigger(t);
				break;
			case 3:
			case 4:
			case 5:
			case 6:
				timers[evtSrc-3].addTrigger(t);
				break;
			case 7:
				addMessageTrigger(t);
				break;
			case 8:
			case 9:
			case 10:
				counters[evtSrc-8].addTrigger(t);
				break;
		}
	}


	/*
	 * Variables
	 */
				
	private void execVariable(int opcode, int var, short value) {
		short newVal = storage.get(var, currentTask.getId());
		
		switch(opcode) {
			case Op.SETV:
				newVal = value;
				break;
			case Op.SUMV:
				newVal += value;
				break;
			case Op.SUBV:
				newVal -= value;
				break;
			case Op.DIVV:
				newVal /= value;
				break;
			case Op.MULV:
				newVal *= value;
				break;
			case Op.SGNV:
				if (newVal < 0)
					newVal = -1;
				else if (newVal > 0)
					newVal = 1;
				else
					newVal = 0;
				break;
			case Op.ABSV:
				if (newVal < 0) newVal = (short) -newVal;
				break;
			case Op.ANDV:
				newVal &= value;
				break;
			case Op.ORV:
				newVal |= value;
				break;
		}
		
		storage.set(var, currentTask.getId(), newVal);
	}
	
	/*
	 * Execution
	 */
	
	protected void enableOpcode(int opcode) {
		opcodeLegal[opcode] = true;
	}
	
	protected void enableOpcodes(int opcodes[]) {
		for(int i=0; i<opcodes.length; ++ i)
			opcodeLegal[opcodes[i]] = true;
	}
	
	private boolean execute() {
		// take care of events
		if (pendingEvents != 0) {
			scheduler.handleEvents(pendingEvents);
			pendingEvents = 0;
		}
		
		Task t = scheduler.next();
		if (t == null) { return false; }
		
		currentTask = t;
		currentPc = t.getPc();
		
		try
		{
			int opcode = t.nextOp();
			
			// if opcode isn't supported on this target, switch to an illegal op
			if (!opcodeLegal[opcode]) {
				opcode = Op.ILLEGAL;
			}
			
			int a1, a2;
			
			switch(opcode)
			{
				// sensors
				case Op.SENT:
					a1 = t.nextByte();
					sensors[a1].setType(t.nextByte(), true);
					break;
				case Op.SENM:
					a1 = t.nextByte();
					sensors[a1].setMode(t.nextByte());
					break;
				case Op.SENZ:
					sensors[t.nextByte()].clear();
					break;
				
				// outputs		
				case Op.OUT:
				case Op.DIR:
				case Op.GDIR:
				case Op.GOUT:
					execOutputs(opcode, t.nextByte(), 0);
					break;
				case Op.PWR:
				case Op.GPWR:
					execOutputs(opcode, t.nextByte(), evaluate(t.nextByte(), t.nextByte()));
					break;
				
				// tasks
				case Op.START:
					start(t.nextByte());
					break;
				case Op.STOP:
					scheduler.stop(t.nextByte());
					break;
				case Op.STOPA:
					stop();
					break;
				case Op.WAIT:
					scheduler.sleep(t.getId(), ticks + evaluate(t.nextByte(), t.nextWord()));
					break;

				
				// flow control
				case Op.JUMP:
				case Op.JUMPL:
					t.setPC(t.nextSignBitTarget(opcode == Op.JUMPL));
					break;
				case Op.CHK:
				case Op.CHKL:
					execCheck(opcode == Op.CHKL);
					break;
				case Op.CALLS:
					t.callSub(codeSpace.getFragmentRange(
						currentProgram, Fragment.SUB_FRAGMENT, t.nextByte()));
					break;
				case Op.RTS:
					t.rts();
					break;
				case Op.LOOPC:
				case Op.LOOPCL:
					execCheckLoop(opcode == Op.LOOPCL);
					break;
				case Op.LOOPS:
					setLoopCounter(evaluate(t.nextByte(), t.nextByte()));
					break;
				case Op.DECVJN:
				case Op.DECVJNL:
					execDecvjn(opcode == Op.DECVJNL);
					break;
					
				// timers
				case Op.TMRZ:
					timers[t.nextByte()].clear();
					break;
				case Op.SETW:
					setWatch(t.nextByte(), t.nextByte());
					break;
					
				// sound
				case Op.PLAYS:
					a1 = t.nextByte();
					if (soundListener != null) soundListener.pbPlaySound(a1);
					break;
				case Op.PLAYT:
					a1 = t.nextWord();
					a2 = t.nextByte();
					if (soundListener != null) soundListener.pbPlayTone(a1, a2);
					break;
				case Op.PLAYV:
					a1 = storage.get(t.nextByte(), t.getId());
					a2 = t.nextByte();
					if (soundListener != null) soundListener.pbPlayTone(a1, a2);					
					break;
				case Op.MUTE:
					if (soundListener != null) soundListener.pbControlSound(SoundListener.MUTE);
					break;
				case Op.SPEAK:
					if (soundListener != null) soundListener.pbControlSound(SoundListener.UNMUTE);
					break;
				case Op.PLAYZ:
					if (soundListener != null) soundListener.pbControlSound(SoundListener.CLEAR);					
					break;
					
				// variables
				case Op.SETV:
				case Op.SUMV:
				case Op.SUBV:
				case Op.DIVV:
				case Op.MULV:
				case Op.SGNV:
				case Op.ABSV:
				case Op.ANDV:
				case Op.ORV:
					execVariable(opcode, t.nextByte(), (short)evaluate(t.nextByte(), t.nextWord()));
					break;
				
				// communication
				case Op.MSG:
					sendMessage(evaluate(t.nextByte(), t.nextByte()));
					break;
				case Op.MSGZ:
					message = 0;
					break;
				case Op.MSGS:
					message = t.nextByte();
					break;

				// display
				case Op.VIEW:
					display.setMode(evaluate(t.nextByte(), t.nextWord()));
					break;
				case Op.DISP:
					t.advance(1);
					display.setUserDisplay(t.nextByte(), t.nextByte(), t.nextWord());
					break;
					
				// datalog
				case Op.LOG:
					datalog.add(t.nextByte(), t.nextByte());
					break;
				case Op.LOGZ:
					datalog.resize(t.nextWord());
					break;
				case Op.POLLD:
					t.nextWord();
					t.nextWord();
					break;
				
				// counters
				case Op.CNTI:
					counters[t.nextByte()].increment();
					break;
				case Op.CNTD:
					counters[t.nextByte()].decrement();
					break;
				case Op.CNTZ:
					counters[t.nextByte()].set((short)0);
					break;
				
				// events
				case Op.SETE:
					setEvent(t.nextByte(), t.nextByte(), t.nextByte());
					break;
				case Op.DELE:
					clearAllEvents();
					break;
				case Op.MONE:
				case Op.MONEL:
					t.enterEvents(evaluate(t.nextByte(), t.nextWord()),
						t.nextSignBitTarget(opcode == Op.MONEL));
					break;
				case Op.MONEX:
					t.exitEvents();
					break;
				case Op.EVENT:
					triggerEvents(evaluate(t.nextByte(), t.nextWord()));
					break;
				
				// resource acccess
				case Op.MONAL:
					t.enterResources(t.nextByte(), t.nextSignBitTarget(true));
					break;
				case Op.MONAX:
					currentTask.exitResources();
					break;
				case Op.SETP:
					currentTask.setPriority(t.nextByte());
					break;
									
				// misc
				case Op.TXS:
				case Op.TOUT:
					t.nextByte();
					break;
				
				case Op.SET:
					setSource(t.nextByte(), t.nextByte(), evaluate(t.nextByte(), t.nextWord()));
					break;

				case Op.ILLEGAL:
					raise(DebugListener.EX_ILLEGAL_OPCODE, opcode, null);
					scheduler.stop(t.getId());
					break;

				default:
					raise(DebugListener.EX_UNKNOWN_OPCODE, opcode, null);
					scheduler.stop(t.getId());
					break;	
			}
		}
		catch (TaskException e) {
			scheduler.stop(t.getId());
		}
		catch (Exception e) {
			raise(DebugListener.EX_JAVA_EXCEPTION, 0, e);
			scheduler.stop(t.getId());		
		}
		
		currentTask = null;
		
		return true;
	}	

	/*
	 * Evaluation
	 */
	int evaluate(int src, int index) {
		try {
			switch(src) {
				case Src.VARIABLE:
					return storage.get(index, currentTaskId());
				case Src.TIMER:
					return timers[index].getTicks();
				case Src.CONSTANT:
					return index;
				case Src.OUTPUT_STATUS:
					return outputs[index].getStatus();
				case Src.RANDOM:
					return (int)(random.nextFloat() * (index + 1));
				case Src.PROGRAM:
					return currentProgram;
				case Src.SENSOR_VALUE:
					return sensors[index].getValue();
				case Src.SENSOR_TYPE:
					return sensors[index].getType();
				case Src.SENSOR_MODE:
					return sensors[index].getMode();
				case Src.SENSOR_RAW:
					return sensors[index].getRawValue();
				case Src.SENSOR_BOOL:
					return sensors[index].getBoolValue();
				
				case Src.WATCH:
					return watch;
					
				case Src.MESSAGE:
					return message;
					
				case Src.GLOBAL_OUTPUT_STATUS:
					return outputs[index].getGlobalStatus();
				
				case Src.COUNTER:
					return counters[index].get();
				case Src.TASK_EVENTS:
					return getActiveEvents(index);
					
				case Src.EVENT_STATE:
					return triggers[index].getState();

				case Src.FAST_TIMER:
					return timers[index].getFast();
				
				case Src.CLICK_COUNTER:
					return triggers[index].getClickCount();
					
				case Src.UPPER_LIMIT:
					return triggers[index].getUpperLimit();
				case Src.LOWER_LIMIT:
					return triggers[index].getLowerLimit();
				case Src.HYSTERESIS:
					return triggers[index].getHysteresis();
				case Src.DURATION:
					return triggers[index].getDuration();
					
				case Src.BATTERY_LEVEL:
					return BATTERY_LEVEL;					
				case Src.FIRMWARE_VERSION:
					return FIRMWARE_VERSION;
					

				default:
					raise(DebugListener.EX_UNKNOWN_SOURCE, src, null);
					return 0;
			}
		}
		catch (Exception e) {
			raise(DebugListener.EX_JAVA_EXCEPTION, 0, e);
			return 0;
		}
	}
	
	
	private int getActiveEvents(int index)
	{
		if (index == scheduler.getTaskCount())
		{
			// use current task
			if (currentTask != null)
				return currentTask.getActiveEvents();
			else
				return 0;
		}

		return scheduler.getTask(index).getActiveEvents();
	}
	
	
	void setSource(int src, int index, int value) {
		try {
			switch(src) {
				case Src.VARIABLE:
					storage.set(index, currentTaskId(), (short)value);
					break;
				case Src.TIMER:
					timers[index].set(value);
					break;

				case Src.RANDOM:
					random.setSeed(value);
					break;
				
				case Src.SENSOR_TYPE:
					sensors[index].setType(value, false);
					break;
				
				case Src.SENSOR_MODE:
					sensors[index].setMode(value);
					break;

				case Src.WATCH:
					setWatch(value/60, value % 60);
					break;

				case Src.MESSAGE:
					deliverMessage(value);
					break;

				case Src.COUNTER:
					counters[index].set((short)value);
					break;
				
				case Src.FAST_TIMER:
					timers[index].setFast(value);
					break;
					
				case Src.CLICK_COUNTER:
					triggers[index].setClickCount(value);
					break;
				case Src.UPPER_LIMIT:
					triggers[index].setUpperLimit(value);
					break;
				case Src.LOWER_LIMIT:
					triggers[index].setLowerLimit(value);
					break;
				case Src.HYSTERESIS:
					triggers[index].setHysteresis(value);
					break;
				case Src.DURATION:
					triggers[index].setDuration(value);
					break;
				
				default:
					raise(DebugListener.EX_UNKNOWN_SOURCE, src, null);
					break;
			}
		}
		catch (Exception e) {
			raise(DebugListener.EX_JAVA_EXCEPTION, 0, e);
		}
	}
	
	private int currentTaskId() {
		if (currentTask != null)
			return currentTask.getId();
		else
			return 0;
	}
	
	void raise(int type, int data, Object o) {
		if (debugListener == null) return;
		
		int task, pc;
		
		if (currentTask != null) {
			task = currentTask.getId();
			pc = currentPc;
		}
		else {
			task = -1;
			pc = -1;
		}
		
		debugListener.pbException(type, task, pc, data, o);
	}
	
};
