/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

import java.io.*;

public class RCXIFormatException extends Exception {
}

/** This class creates instances of Programs from NQC output
 *	files.
 */
public class RCXIReader{

	private		InputStream s;

	private static final int RCXI_SIGNATURE = 0x49584352;
	private static final int RCXI_VERSION = 0x101;

	/** Instantiate a new reader for the given InputStream
	*	@param is	InputStream of an RCXI formatted file (such as produced by NQC).
	*/
	public	RCXIReader(InputStream is) { s = is; }
	
	/** Read the file and create a Program object.
	*	Note it is an error to call this method more than once for a given
	*	RCXIReader.
	*	@returns	a program object suitable for a call to PBrick.load()
	*/
	public Program readProgram() throws IOException, RCXIFormatException {
		// check signature and version
		if (get4() != RCXI_SIGNATURE) throw new RCXIFormatException();
		if (get2() < RCXI_VERSION) throw new RCXIFormatException();

		// get counts
		int fragmentCount = get2();
		int symbolCount = get2();		

		// skip reserved
		s.skip(2);
		
		// read fragments
		Program p = new Program(fragmentCount);
		
		for(int i=0; i<fragmentCount; ++i)
		{
			int type = get1();
			int index = get1();
			int length = get2();
			
			byte data[] = new byte[length];
			s.read(data);
			
			// pad
			int pad = (4 - length) & 3;
			s.skip(pad);
			
			int fType;
			
			// Currently RCXI types are the same as PBrick types,
			// but we'll decode them anyway just to be safe in case
			// the PBrick constants change later on
			if (type == 0)
				fType = Fragment.TASK_FRAGMENT;
			else
				fType = Fragment.SUB_FRAGMENT;
			
			p.setFragment(i, new Fragment(fType, index, data));
		}
		
		return p;
	}
	
	
	private int		get4() throws IOException {
		int v = get1();
		v += get1() << 8;
		v += get1() << 16;
		v += get1() << 24;
		
		return v;
	}
	
	private int		get2() throws IOException {
		int v = get1();
		v += get1() << 8;

		return v;
	}
	
	private int		get1() throws IOException {
		int b = s.read();
		if (b==-1) throw new IOException();
		return b;
	}
}
