/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

class Scheduler {
	private Task			tasks[];	
	private TaskQueue		runningTasks;
	private	TaskQueue		idleTasks;
	private TaskQueue		sleepingTasks;
	private Task			nextTask;
	private DebugListener	debugListener;
	private PBrick			pb;
	
	Scheduler(PBrick pb, int taskCount, byte code[]) {
		this.pb = pb;
		
		runningTasks = new TaskQueue(Task.RUNNING);
		idleTasks = new TaskQueue(Task.IDLE);
		sleepingTasks = new TaskQueue(Task.SLEEPING);
		
		tasks = new Task[taskCount];
		for(int i=0; i<taskCount; ++i) {
			tasks[i] = new Task(this, i, idleTasks, code);
		}
				
		nextTask = null;
	}
	
	void setDebugListener(DebugListener l) { debugListener = l; }
	
	Task getTask(int i) { return tasks[i]; }
	int getTaskCount() { return tasks.length; }
	
	
	boolean	isRunning() {
		if (runningTasks.getHead() != null) return true;
		if (sleepingTasks.getHead() != null) return true;
		return false;
	}
	
	Task next() {
		Task t = nextTask;
		if (t != null) {
			nextTask = t.getNext();
			if (nextTask == null) nextTask = runningTasks.getHead();
		}
		return t;
	}
	
	/** stop all running and sleeping tasks
	 */
	void stopAll() {
		for(int i=0; i<tasks.length; ++i)
			stop(i);
	}
	
	/** stop the designated task.  Do not call the task's stop method
	 *	directly - call this method instead.
	 */
	void stop(int id) {
		Task t = tasks[id];
		if (t.getQueue() == idleTasks) return;
		
		avoidNextTask(t);
		t.enqueue(idleTasks);
		t.stop();
		
		if (debugListener != null)
			debugListener.pbTaskChanged(t);
	}


	/** start the specified task executing in the given range of
	 *	addresses.  Do not call the task's start method directly,
	 *	use this method instead
	 */
	void start(int id, Range r) {
		Task t = tasks[id];
		
		if (nextTask == null) nextTask = t;

		t.enqueue(runningTasks);
		t.start(r);

		if (debugListener != null)
			debugListener.pbTaskChanged(t);
	}
	
	
	void sleep(int id, int time) {
		Task t = tasks[id];
		avoidNextTask(t);
		t.setWakeTime(time);
		t.enqueue(sleepingTasks);

		if (debugListener != null)
			debugListener.pbTaskChanged(t);
	}
	

	void wakeup(int ticks) {
		// look for tasks to wake up
		Task t, next;
		for(t = sleepingTasks.getHead(); t!=null; t=next) {
			next = t.getNext();
			if (t.getWakeTime() <= ticks) {	
				wakeup(t);
			}
		}
	}


	void wakeup(Task t) {
		if (t.getQueue() != sleepingTasks) return;
		
		t.enqueue(runningTasks);
		if (nextTask == null) nextTask = t;
		if (debugListener != null)
			debugListener.pbTaskChanged(t);
	}


	void handleEvents(int events) {
		Task t, next;
		
		for(t = runningTasks.getHead(); t!= null; t = t.getNext()) {
			t.handleEvents(events);
		}

		for(t = sleepingTasks.getHead(); t!= null; t=next) {
			next = t.getNext();
			if (t.handleEvents(events)) {
				wakeup(t);
			}
		}		
	}


	/** attempt to acquire the specified resources at the given priority.
	 *	This may case resources to be lost by other tasks
	 */
	boolean	acquire(int resources, int priority) {
		
		// first check to see if possible
		for(int i=0; i<tasks.length; ++i) {
			Task t = tasks[i];
			if (((t.getResources() & resources) != 0) &&
				(t.getPriority() < priority))
					return false;
		}
		
		// now free the resources
		for(int i=0; i<tasks.length; ++i) {
			Task t = tasks[i];
			int contended = t.getResources() & resources;
			if (contended != 0) {
				for(int j=0; j<pb.MAX_RESOURCES; ++j)
					if ((contended & (1 << j)) != 0)
						pb.releaseResource(j);
				t.loseResources();
			}
		}
		return true;
	}
	
	
	private void avoidNextTask(Task t) {
		if (t != nextTask) return;

		// find the next task (treating queue as circular)
		nextTask = t.getNext();
		if (nextTask == null) nextTask = runningTasks.getHead();
		
		// if we're back at t, then there is no next task
		if (nextTask == t) nextTask = null;
	}
}