/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

public class Sensor {
	public static final int MAX_RAW_VALUE = 1023;
	public static final int	MAX_CUMULATIVE_VALUE = 32767;
	public static final int MIN_CUMULATIVE_VALUE = -32768;
	public static final int ILLEGAL_TEMP_VALUE = 32767;

	// sensor types - MUST MATCH RCX DEFINITIONS
	public static final int TYPE_NONE = 0;
	public static final int TYPE_TOUCH = 1;
	public static final int TYPE_TEMPERATURE = 2;
	public static final int TYPE_LIGHT = 3;
	public static final int TYPE_ROTATION= 4;
	
	// sensor modes - MUST MATCH RCX DEFINITIONS
	public static final int MODE_RAW = 0x00;
	public static final int MODE_BOOL = 0x20;
	public static final int MODE_EDGE = 0x40;
	public static final int MODE_PULSE = 0x60;
	public static final int MODE_PERCENT = 0x80;
	public static final int MODE_CELSIUS = 0xa0;
	public static final int MODE_FAHRENHEIT = 0xc0;
	public static final int MODE_ROTATION = 0xe0;
	public static final int MODE_MASK = 0xe0;
	public static final int SLOPE_MASK = 0x1f;
	
	private int id;
	private int mode;
	private int type;
	
	private int pendingValue;
	private boolean pendingCumulative;
	private int	rawValue;
	private int value;
	private boolean boolValue;
	private int lastQuad;	// only used in MODE_ROTATION
	
	private SensorListener listener;
	private TriggerQueue triggers;
	private PBrick pb;
	

	private static final int defaultModes[] = {
		MODE_RAW, MODE_BOOL, MODE_CELSIUS,
		MODE_PERCENT, MODE_ROTATION
	};

	
	Sensor(PBrick pb, int id) {
		this.pb = pb;
		this.id = id;
		listener = null;
		triggers = new TriggerQueue();
		reset();
	}
	
	public int getId() 		{ return id; }
	
	public int 	getValue()		{ return value; }
	public int 	getRawValue()	{ return rawValue; }
	public int	getBoolValue()	{ return boolValue ? 1 : 0; }
	
	public int	getMode()		{ return mode; }
	public int	getType()		{ return type; }


	public void setListener(SensorListener l) { listener = l; }
	
	void addTrigger(Trigger t) {
		t.enqueue(triggers);
		
		int low = 0;
		int hi = PBrick.MAX_INT;
		
		switch(mode) {
			case MODE_RAW:
				hi = 1023;
				break;
			case MODE_BOOL:
				hi = 1;
				break;
			case MODE_PERCENT:
				hi = 100;
				break;
			case MODE_CELSIUS:
				low = -200;
				hi = 700;
				break;
			case MODE_FAHRENHEIT:
				low = -40;
				hi = 1580;
				break;
			case MODE_ROTATION:
				low = PBrick.MIN_INT;
				break;
		}
		
		t.prime(value, low, hi);
	}
	
	
	private void reset() {
		mode = MODE_RAW;
		type = TYPE_NONE;
		pendingValue = rawValue = value = 1023;
		boolValue = false;
		lastQuad = rawToQuadrant(rawValue);
		pendingCumulative = false;
	}
	
	
	/** set the raw value of a sensor.  The new value won't take effect
	 *	until the next time the sensor is sampled (normally every time
	 *	PBrick.step() is called).
	 */
	public synchronized void setRawValue(int v) {
		if (v > MAX_RAW_VALUE)
			v = MAX_RAW_VALUE;
		else if (v < 0)
			v = 0;
			
		pendingValue = v;
		pendingCumulative = false;
		pb.setSampleSensors();
	}

	/** set the processed value of a sensor that is in a cumulative mode
	 *	(MODE_ROTATION, MODE_EDGE, or MODE_PULSE).  This call is especially
	 *	convenient when working with rotation sensors since sending the
	 *	correct sequence of raw values to represent rotation through 
	 *	quadrants is quite tedious.  The new value does not take effect
	 *	until the next time the sensor is sampled.
	 */
	public synchronized void setCumulativeValue(int v) {
		if (!isCumulative()) return;

		if (v > MAX_CUMULATIVE_VALUE)
			v = MAX_CUMULATIVE_VALUE;
		else if (v < MIN_CUMULATIVE_VALUE)
			v = MIN_CUMULATIVE_VALUE;

		pendingCumulative = true;
		pendingValue = v;
		pb.setSampleSensors();
	}
		
	
	synchronized void sample() {
		if (pendingCumulative) {
			value = pendingValue;
			pendingValue = rawValue;
			pendingCumulative = false;
			triggers.update(value, 0);
		} else if (rawValue != pendingValue) {
			processRawValue(pendingValue);
		}
	}
	
	
	/** convert a percent back to a raw value.  This method can be used to 
	 *	determine a raw sensor value that will yield the specified percentage
	 *	when a sensor is in SENSOR_MODE_PERCENT mode.
	 *	@param percent desired percentage
	 *	@returns raw value
	 */
	public static int percentToRaw(int percent) {
		/* this may look like a bug since a value of
		 * 101 is allowed but anything larger gets clipped to 100
		 * but that's just to be consistent with a bug in the firmware
		 * percentage conversion (which is faithfully reproduced here
		 */
		if (percent > 101)
			percent = 100;
		else if (percent < 0)
			percent = 0;
			
		/* The conversion from percent to raw adds 24 to the
		 * numerand before dividing by 25 in order to preserve stability
		 * of percent -> raw -> percent conversions.  If anyone is curious
		 * about the general formula, given a conversion in integer math of
		 *     f(x) = (x * a) / b
		 * if you want to find an inverse g(x) such that
		 *     f(g(x)) = x
		 * then as long as a <= b,
		 *     g(x) = (x * b + (a - 1)) / a
		 * basically, the (a-1) term effectively makes /a rounded up to 
		 * compensate for f(x) rounding down.  Note that if b > a, then
		 * for some values y, no x exists such that f(x) = y, therefore
		 * no g(y) could exist such that f(g(y)) = y
		 *
		 * quite a long-winded comment for one line of code
		 */
		int raw = 1023 - (percent * 156 + 24) / 25;
		
		return raw;
	}
	
	private static final int SEGMENT_COUNT = 7;
	
	private static int temperatureSegments[] = {
		290, -261, 1452,
		340, -194, 1230,
		390, -144, 1035,
		478, -123, 936,
		520, -107, 852,
		790, -134, 1066,
		886, -175, 1426,
		928
	};
	
	
	private static int rawToCelsius(int raw) {
		// find segment
		int seg;
		for(seg=0; seg<SEGMENT_COUNT+1; ++seg)
			if (raw < temperatureSegments[seg*3]) break;

		if (seg==0 || seg > SEGMENT_COUNT) return ILLEGAL_TEMP_VALUE;
		
		int scale = temperatureSegments[seg*3-2];
		int offset = temperatureSegments[seg*3-1];
		
		return (raw * scale) / 100 + offset;
	}

	
	void setMode(int m) {
		if (mode == m) return;
		mode = m;
		clear();
		lastQuad = rawToQuadrant(rawValue);
		processRawValue(rawValue);
		if (listener != null) listener.pbSensorChanged(this);
	}

	
	void setType(int t, boolean setDefaultMode) {
		type = t;
		
		if (setDefaultMode)
			setMode(defaultModes[t]);
		else {
			processRawValue(rawValue);
			if (listener != null) listener.pbSensorChanged(this);
		}
	}


	void clear() {
		if (isCumulative())
			value = 0;
	}


	private boolean isCumulative() {
		switch(mode) {
			case MODE_EDGE:
			case MODE_PULSE:
			case MODE_ROTATION:
				return true;
			default:
				return false;
		}
	}
	
	private void processRawValue(int raw) {
		int v = value;
		boolean b = computeBoolValue(raw);
		int external = 0;
		int newQuad;
		
		rawValue = raw;
		
		switch(mode & MODE_MASK) {
			case MODE_RAW:
				v = rawValue;
				break;
			case MODE_BOOL:
				v = b ? 1 : 0;
				break;
			case MODE_EDGE:
				if (b != boolValue) {
					external |= 1 << Trigger.EVENT_TYPE_EDGE;
					++v;
				}
				break;
			case MODE_PULSE:
				if (b != boolValue && (b)) {
					external |= 1 << Trigger.EVENT_TYPE_PULSE;
					++v;
				}
				break;
			case MODE_PERCENT:
				// due to a bug in the RCX ROM, the percent conversion can
				// actually return a value of 101...this bug is implemented
				// faithfully in the simulator
				if (raw < 391)
					v = 100;
				else
					v = (1023 - raw) * 25 / 156;
				break;			
			case MODE_CELSIUS:
				v = rawToCelsius(raw);
				break;
			case MODE_FAHRENHEIT:
				v = rawToCelsius(raw);
				if (v != ILLEGAL_TEMP_VALUE)
					v = v * 9 / 5 + 320;
				break;
			case MODE_ROTATION:
				newQuad = rawToQuadrant(raw);
				if (newQuad != lastQuad) {
					if (newQuad == ((lastQuad + 1) & 3))
						++v;
					else if (newQuad == ((lastQuad - 1) & 3))
						--v;
					lastQuad = newQuad;
				}
				break;
		}

		// return if no meaningful change
		if (boolValue == b && value == v) return;

		// pressed, released, and fastchange events
		if (boolValue != b) {
			// button was either pressed or released
			external |= 1 << (b ? Trigger.EVENT_TYPE_PRESSED : Trigger.EVENT_TYPE_RELEASED);

			// check for fast change (anytime a change takes place in slope mode)
			if ((mode & SLOPE_MASK) != 0)
				external |= 1 << Trigger.EVENT_TYPE_FASTCHANGE;
		}
		
		boolValue = b;
		value = v;

		triggers.update(v, external);
	}	


	private boolean computeBoolValue(int raw) {
		int slope = mode & SLOPE_MASK;
		
		if (slope == 0) {
			// no slope, use 45%/55% cutoffs
			if (raw < 0x1cc)
				return true;
			else if (raw > 0x232)
				return false;

		} else {
			// slope mode - first check ends
			if (raw > MAX_RAW_VALUE - slope)
				return false;
			else if (raw < slope)
				return true;

			// now check rate of change
			int delta = raw - rawValue;
			
			if (delta > slope)
				return false;
			else if (delta < -slope)
				return true;
		}
		
		// no change
		return boolValue;
	}


	/** convert a raw value into a quadrant. */
	private static int rawToQuadrant(int raw) {
		if (raw <= 0x1bf)
			return 0;
		else if (raw <= 0x2bf)
			return 1;
		else if (raw <= 0x3bf)
			return 3;
		else
			return 2;
	}
}
