/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

/** This class provides an easy way to run the simulation on its own thread.
*	The thread can be have one of three speeds:
*	<p>
*	PAUSED - simulation is halted
*	<p>
*	NORMAL - simulation will attempt to run in real time
*	<p>
*	FAST - simulation will run as fast as it can.  This can be hundreds of 
*	times faster than an actual RCX.
*/

public class SimThread extends Thread {
	public static final int PAUSED = 0;
	public static final int NORMAL = 1;
	public static final int FAST = 2;
	
	// time slicing param - smaller is more accurate, larger is
	// more efficient.
	private static final int SLICE = 10; 

	private PBrick pb;
	private int speed = PAUSED;
	
	
	/** Create a simulation thread for the specified PBrick.  Like all threads,
	*	this thread must be started by calling start().
	*	@param pb	the PBrick to be simulated
	*/
	public SimThread(PBrick pb) {
		this.pb = pb;
	}

	
	/** Change the speed of the simulation thread.
	*	@param speed new speed (PAUSED, NORMAL, or FAST)
	*/
	public void setSpeed(int speed)	{
		if (speed==this.speed) return;
		
		if (this.speed == FAST) {
			// just switch the mode, the loop will notice real soon
			this.speed = speed;
		}
		else {
			// need to synchronize and notify the thread
			synchronized(this) {
				this.speed = speed;
				notify();
			}
		}
	}
	
	/** Get the current speed of the simulation thread.
	*	@returns	current speed (PAUSED, NORMAL, or FAST)
	*/
	public int getSpeed() { return speed; }

	
	public synchronized void run() {
		while(true) {
			try {
				if (speed == FAST) {
					while(speed == FAST) {
						pb.step();
					}
				}
				else if (speed == NORMAL) {
					long baseTime = System.currentTimeMillis() - pb.getTicks() * 10;
					
					while(speed == NORMAL) {
						// run at normal speed
						pb.runFor(SLICE);
						long delta = baseTime + pb.getTicks() * 10 - System.currentTimeMillis();
						if (delta > 0) {
							wait(delta);
						}
					}
				}
				else {
					wait();
				}
			}
			catch (InterruptedException e) {
				// don't do anything - this is expected when switching out of
				// PAUSED and NORMAL modes
			}
		}
	}
}
