/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

class TaskException extends Exception {
	public static final int BOUNDS = 0;
	public static final int SUBROUTINE = 1;
	
	private int	reason;

				TaskException(int r) { reason = r; }
	
	int getReason() { return reason; }
}


class TaskQueue
{
	Task	head;
	int		state;
	
	TaskQueue(int state) {
		this.state = state;
		head = null;
	}
	
	Task	getHead() { return head; }
}	


public class Task
{
	// values for getState()
	public static final int IDLE = 0;
	public static final int RUNNING = 1;
	public static final int SLEEPING = 2;
	
	private Scheduler scheduler;
	private int id;

	// scheduling
	private Task		next;
	private TaskQueue	queue;
	private int			wakeTime;

	// basic environment
	private byte			code[];

	// current execution information
	private int				pc;
	private int				start;
	private int				end;
	
	// saved state prior to subroutine call
	private static final int NO_RETURN = -1;
	private int				returnPC = NO_RETURN;
	private int				returnStart;
	private int				returnEnd;
	
	// event monitoring
	private int				monitoredEvents = 0;
	private int				activeEvents;
	private int				eventPC;
	
	// resource access
	private int				priority = 0;
	private	int				resources = 0;
	private int				resourcePC;
	
	// event/resource nesting
	private boolean			evtMostRecent;
	
	Task(Scheduler scheduler, int id, TaskQueue q, byte code[]) {
		this.scheduler = scheduler;
		this.id = id;
		this.code = code;
		pc = 0;
		insert(q);
	}
	
	public int 	getId() 	{ return id; }
	public int	getState()	{ return queue.state; }

	public int getActiveEvents() { return activeEvents; }
	
	Task	getNext()	{ return next; }
	int		getWakeTime()	{ return wakeTime; }
	void	setWakeTime(int w)	{ wakeTime = w; }
	TaskQueue	getQueue() 	{ return queue; }
	int		getPc()		{ return pc; }

	int	getPriority() { return priority; }
	void setPriority(int p) { priority = p; }
	int getResources() { return resources; }
	
	int nextOp() throws TaskException {
		// very similar to nextByte(), except will return from subroutines
		if (pc >= end && returnPC != NO_RETURN)
			rts();

		if (pc >= end)
			throw new TaskException(TaskException.BOUNDS);
		
		return (int)code[pc++] & 0xff;
	}
	
	void rts() throws TaskException {
		if (returnPC == NO_RETURN) throw new TaskException(TaskException.SUBROUTINE);
		pc = returnPC;
		start = returnStart;
		end = returnEnd;
		returnPC = NO_RETURN;
	}
	
	
	void advance(int n) throws TaskException {
		if (pc + n > end) throw new TaskException(TaskException.BOUNDS);
		pc += n;
	}
	
	int	nextByte() throws TaskException {
		if (pc >= end) throw new TaskException(TaskException.BOUNDS);
		
		return (int)code[pc++] & 0xff;
	}
	
	
	int nextWord() throws TaskException {
		int v = nextByte();
		v += nextByte() << 8;
		return v;
	}
	
	
	int nextSignBitTarget(boolean longJump) throws TaskException {
		int old = pc;
		int b1 = nextByte();		
		int delta = (b1 & 0x7f);
		
		if (longJump)
			delta += (nextByte() << 7);
		
		// direction bit
		if ((b1 & 0x80) != 0)
			delta = -delta;
		
		return old + delta;
	}


	int nextOffsetTarget(boolean longJump) throws TaskException {
		int old = pc;
		short delta;
		
		if (longJump)
			delta = (short) nextWord();
		else
			delta = (byte) nextByte();

		return old + delta;
	}


	void setPC(int pc) {
		this.pc = pc;
		if (pc < start)
			pc = end;
	}
	

	void start(Range r) {
		start = pc = r.getStart();
		end = r.getEnd();
		returnPC = NO_RETURN;
	}
	
	
	void stop() {
		monitoredEvents = 0;
		resources = 0;
	}


	void callSub(Range r) throws TaskException {
		if (returnPC != NO_RETURN) {
			// cannot call sub from within sub
			throw new TaskException(TaskException.SUBROUTINE);
		}
		
		returnStart = start;
		returnEnd = end;
		returnPC = pc;
		
		start = pc = r.getStart();
		end = r.getEnd();
	}
	
	
	void enterEvents(int events, int addr) {
		activeEvents = 0;
		monitoredEvents = events;
		eventPC = addr;
		evtMostRecent = true;
	}
	
	void exitEvents() {
		monitoredEvents = 0;
		
		// see if resources nested within events
		if ((resources != 0) && !evtMostRecent) {
			resources = 0;
		}
	}
	
	
	boolean handleEvents(int events) {
		if (monitoredEvents == 0) return false;
		
		activeEvents = events & monitoredEvents;
		if (activeEvents != 0) {
			monitoredEvents = 0;
			pc = eventPC;
			return true;
		}
		
		return false;
	}


	void enterResources(int resources, int addr) {
		this.resources = 0;

		evtMostRecent = false;
		if (scheduler.acquire(resources, priority)) {
			resourcePC = addr;
			this.resources = resources;
		}
		else {
			pc = addr;
		}
	}
	
	
	void exitResources() {
		resources = 0;
		
		// see if events nested within resources
		if ((monitoredEvents != 0) && evtMostRecent) {
			monitoredEvents = 0;
		}
	}


	void loseResources() {
		exitResources();
		pc = resourcePC;
		scheduler.wakeup(this);
	}

	
	void enqueue(TaskQueue q) {
		if (q==queue) return;
		remove();
		insert(q);
	}


	private void insert(TaskQueue q) {
		next = q.head;
		q.head = this;
		queue = q;
	}
	
	
	private void remove() {

		if (queue==null) return;
		
		// simple case of task at head
		if (queue.head == this)
		{
			queue.head = this.next;
			return;
		}
		
		// find previous
		Task prev = null;
		for(Task t=queue.head; t!=null; t=t.next) {
			if (t == this) {
				prev.next = this.next;
				return;
			}
			prev = t;
		}
		
		// task not in the queue! do nothing
	}
};
