/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbsim;

class TriggerQueue {
	Trigger head;
	
	Trigger	getHead() { return head; }
	
	void update(int v, int external) {
		for(Trigger t = head; t!=null; t=t.getNext()) {
			t.update(v, external);
		}
	}
}


class Trigger {
	// event types
	public static final int EVENT_TYPE_PRESSED = 0;
	public static final int EVENT_TYPE_RELEASED = 1;
	public static final int EVENT_TYPE_PULSE = 2;
	public static final int EVENT_TYPE_EDGE = 3;
	public static final int EVENT_TYPE_FASTCHANGE = 7;
	public static final int EVENT_TYPE_LOW = 8;
	public static final int EVENT_TYPE_NORMAL = 9;
	public static final int EVENT_TYPE_HIGH = 10;
	public static final int EVENT_TYPE_CLICK = 11;
	public static final int EVENT_TYPE_DOUBLECLICK = 12;
	public static final int EVENT_TYPE_MESSAGE = 14;

	// internal state
	private static final int STATE_LOW = 0;
	private static final int STATE_NORMAL = 1;
	private static final int STATE_HIGH = 2;
	private static final int STATE_UNDEFINED = 3;
	private static final int STATE_START_CAL = 4;
	private static final int STATE_CALIBRATING = 5;

	// useful constants
	private static final int DEFAULT_DURATION = 15;
	
	// queue management
	private TriggerQueue	queue;
	private Trigger next;
	
	// event wiring
	private PBrick pb;
	private int id;
	private int type;
	
	// current value and state
	private int value;
	private int state = STATE_UNDEFINED;
	private short clickCount = 0;
	private int clickState;
	private int lastTransition;
	
	
	// parameters - could change this to an array of ints since
	// there are get/set functions for each.  However, this would
	// make the event checking code harder to read and slower to
	// execute.
	private int	lowerLimit;
	private int upperLimit;
	private int hysteresis;
	private int duration;	
	
	TriggerQueue	getQueue()	{ return queue; }
	Trigger			getNext() 	{ return next; }
	

	Trigger(PBrick pb, int id) {
		this.pb = pb;
		this.id = id;
	}
	
	void setType(int type) {
		this.type = type;
	}
	
	int getState() { return state; }
	int getClickCount() { return clickCount; }
	
	int getUpperLimit()	{ return upperLimit; }
	int getLowerLimit() { return lowerLimit; }
	int getHysteresis() { return hysteresis; }
	int getDuration() { return duration; }
	
	void setUpperLimit(int x) {
		upperLimit = x;
		check();
	}
		
	void setLowerLimit(int x) {
		lowerLimit = x;
		check();
	}
	
	void setHysteresis(int x) {
		hysteresis = x;
		check();
	}
	
	void setDuration(int x) {
		duration = x;
		check();
	}
	
	
	void setClickCount(int x) {
		clickCount = (short)x;
	}
	
	void update(int v, int external) {
		value = v;
		check();
		
		// handle the case of an external trigger
		if (((1 << type) & external) != 0)
			pb.triggerEvents(1 << id);
	}
	
	
	void prime(int v, int low, int high) {
		lowerLimit = low;
		upperLimit = high;
		value = v;
		duration = DEFAULT_DURATION;
		clickState = 0;
		clickCount = 0;
		state = computeState(v);
	}
	
	
	private int checkClicks(int newState) {
		// don't count transitions that don't involve STATE_HIGH
		if (state!=STATE_HIGH && newState != STATE_HIGH) return 0;
		
		// update transition time
		int thisTrans = pb.getTicks();
		boolean inTime = (thisTrans-lastTransition <= duration);
		lastTransition = thisTrans;
		
		// state machine
		switch(clickState) {
			case 1:
				if (inTime) {
					clickState = 2;
					clickCount++;
					return 1;
				}
				break;
			case 2:
				if (inTime) {
					clickState++;
				}
				return 0;
			case 3:
				if (inTime) {
					clickState = 0;
					clickCount++;
					return 2;
				}
				break;
		}
		
		// default is 0 or 1 depending on newState
		clickState = (newState == STATE_HIGH) ? 1 : 0;
		return 0;
	}
	
	
	private void check() {
		int newState = computeState(value);
		
		// return if no change in state
		if (state == newState) return;
		
		// check for clicks
		int recentClicks = checkClicks(newState);
				
		state = newState;		
		boolean t = false;
		
		switch(type) {
			case EVENT_TYPE_LOW:
			case EVENT_TYPE_NORMAL:
			case EVENT_TYPE_HIGH:
				t = (state == (type - EVENT_TYPE_LOW));
				break;
			case EVENT_TYPE_CLICK:
				t = (recentClicks > 0);
				break;
			case EVENT_TYPE_DOUBLECLICK:
				t = (recentClicks > 1);
				break;
		}
				
		if (t)
			pb.triggerEvents(1 << id);
	}
	
	
	private int computeState(int v) {
		int cutoff;
		
		// check upper limit
		cutoff = upperLimit;
		if (state == STATE_HIGH)
			cutoff -= hysteresis;
		if (v > cutoff)
			return STATE_HIGH;
		
		// check lower limit
		cutoff = lowerLimit;
		if (state == STATE_LOW)
			cutoff += hysteresis;
		if (v < cutoff)
			return STATE_LOW;
		
		return STATE_NORMAL;
	}

	
	void enqueue(TriggerQueue q) {
		if (q==queue) return;
		remove();
		insert(q);
	}
	

	private void insert(TriggerQueue q) {
		if (q==null) {
			queue = null;
			next = null;
		}
		else {
			next = q.head;
			q.head = this;
			queue = q;
		}
	}

	
	private void remove() {
		if (queue == null) return;
		
		// simple case of task at head
		if (queue.head == this)
		{
			queue.head = this.next;
			return;
		}
		
		// find previous
		Trigger prev = null;
		for(Trigger t=queue.head; t!=null; t=t.next) {
			if (t == this) {
				prev.next = this.next;
				return;
			}
			prev = t;
		}
		
		// task not in the queue! do nothing
	}
}
