/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

package pbtest;

import java.io.*;
import pbsim.*;
import java.awt.*;
import java.awt.event.*;


public class RCXTest implements OutputListener, DebugListener, SoundListener
{
	private static final String outModeNames[] = { "FLT", "OFF", "ON " };
	private static final String outDirNames[] = { "REV", "FWD" };
	
	public static void main(String args[]) {

		// process command line
		if (args.length != 1) {
			System.out.println("Usage: RCXSim filename");
			return;
		}
		String filename = args[0];
		
		// try to read RCX program
		Program p;
		try {
			InputStream stream = new FileInputStream(filename);
			RCXIReader r = new RCXIReader(stream);
			p = r.readProgram();
		} catch (Exception e) {
			System.out.println("Error: could not read RCX program from " + filename);
			return;
		}
		
		new RCXTest(p);
	}
	
	private RCX rcx;
	
	private	RCXTest(Program p) {
		rcx = new RCX();

		rcx.setOutputListener(this);
		rcx.setSoundListener(this);
		rcx.setDebugListener(this);
		
		RCXFrame f = new RCXFrame(rcx);
		f.show();
		
		rcx.load(p);
		rcx.start();
		f.setSpeed(SimThread.NORMAL);
	}
	
	
	// callbacks for all of the listeners
	
	public void pbOutputChanged(Output p) {
		log("Output " + (char)('A'+p.getId()) + " " +
			outModeNames[p.getMode()] + " " +
				(p.getForward() ? "Fwd" : "Rev") + " " + p.getPower());
	}
	
	public void pbTaskChanged(Task t) {
		log("Task " + t.getId() + " -> " + t.getState());
	}
	
	public void pbPlaySound(int sound) {
		log("Sound " + sound);
	}
	
	public void pbPlayTone(int freq, int duration) {
		log("Tone " + freq + ", " + duration);
	}
	
	public void pbControlSound(int code) {
		log("Sound control " + code);
	}
		
	public void pbException(int type, int task, int pc, int data, Object o) {
		System.out.print("pbException @" + pc + " (" + task + ") :");
		switch(type) {
			case EX_ILLEGAL_OPCODE:
				System.out.println("Illegal opcode 0x" + Integer.toHexString(data));
				break;
			case EX_UNKNOWN_OPCODE:
				System.out.println("Unknown opcode 0x" + Integer.toHexString(data));
				break;
			case EX_UNKNOWN_SOURCE:
				System.out.println("Unknown source " + data);
				break;
			case EX_JAVA_EXCEPTION:
				System.out.println(o);
				break;
		}
	}
	
	private void log(String s) {
		System.out.println(rcx.getTicks() + ": " + s);
	}
}


class RCXFrame extends Frame {

	private RCX rcx;
	RunControls runControls;
	
	public RCXFrame(RCX rcx) {
		super("RCX Simulator");
		
		this.rcx = rcx;
		setLayout(new GridLayout(rcx.getSensorCount() + 3, 1));
		
		add(new LCD(rcx.getDisplay()));
		
		for(int i=0; i<rcx.getSensorCount(); ++i)
			add(new SensorControl(rcx.getSensor(i)));
		
		add(new MessageCenter(rcx));
		
		runControls = new RunControls(rcx);
		add(runControls);
	}
	
	public void setSpeed(int speed) { runControls.setSpeed(speed); }
}



class LCD extends Canvas implements DisplayListener {

	private Display display;
	private Font	font;
	char			chars[];

	public LCD(Display d) {
		font = Font.decode("Monospaced-BOLD-18");
		setFont(font);

		display = d;
		setSize(100, 20);
		display.setListener(this);
		
		chars = new char[Display.WIDTH+1];
	}
	
	public void paint(Graphics g) {
		int i;
		for(i=0; i<Display.WIDTH - display.getPoint(); ++i)
			chars[i] = display.getChar(i);
			
		if (display.getPoint() != 0) {
			chars[i] = '.';
			for( ; i<Display.WIDTH; ++i)
				chars[i+1] = display.getChar(i);
			++i;
		}
			
		g.drawChars(chars, 0, i, 20, 20);
	}
	
	public void pbDisplayValueChanged(Display d) {
		repaint();
	}
	
	public void pbDisplayModeChanged(Display d) {
	}
};


class RunControls extends Container implements ActionListener, ItemListener {
	
	private PBrick pb;
	Choice	speedChoice;
	SimThread thread;
	
	public RunControls(PBrick pb) {
		this.pb = pb;
		
		thread = new SimThread(pb);
		thread.start();
		
		setLayout(new FlowLayout(FlowLayout.LEFT));
		
		Button b;
		
		b = new Button("Start");
		b.addActionListener(this);
		add(b);

		b = new Button("Stop");
		b.addActionListener(this);
		add(b);
		
		speedChoice = new Choice();
		speedChoice.add("Paused");
		speedChoice.add("Normal");
		speedChoice.add("Fast");
		speedChoice.addItemListener(this);
		add(speedChoice);
	}
	
	
	public void setSpeed(int m) {
		speedChoice.select(m);
		thread.setSpeed(m);
	}
	
	public void actionPerformed(ActionEvent evt) {
		String cmd = evt.getActionCommand();
		
		if (cmd == "Start") {
			pb.start();
		}
		else if (cmd == "Stop") {
			pb.stop();
		}
	}
	
	public void itemStateChanged(ItemEvent evt) {
		thread.setSpeed(speedChoice.getSelectedIndex());
	}
}


class MessageCenter extends Container implements ActionListener, MessageListener {
	PBrick 		pb;
	Label		outMsg;	// from RCX to world
	TextField	inMsg;	// from world to RCX
	
	public MessageCenter(PBrick pb) {
		this.pb = pb;
		
		setLayout(new FlowLayout(FlowLayout.LEFT));
		add(new Label("Incoming Message:"));
		add(inMsg = new TextField("0", 3));
		add(new Label("Outgoing Message:"));
		add(outMsg = new Label("       "));
		
		pb.setMessageListener(this);
		inMsg.addActionListener(this);
	}
	

	public void pbMessageSent(int m) {
		outMsg.setText(String.valueOf(m));
	}
	
	public void actionPerformed(ActionEvent evt) {
		try {
			int v = Integer.parseInt(inMsg.getText());
			pb.deliverMessage(v);
		}
		catch (Exception e) {
			inMsg.setText("");
		}
	}
}


class SensorControl extends Container implements ItemListener, AdjustmentListener, ActionListener, SensorListener {
	private Sensor sensor;
	private Checkbox	checkbox;
	private Scrollbar	scrollbar;
	private TextField	textfield;
	private Component	active;
	
	
	public SensorControl(Sensor s) {
		s.setListener(this);
		sensor = s;

		setLayout(new FlowLayout(FlowLayout.LEFT));
		add(new Label("Sensor " + String.valueOf(s.getId() + 1)));
		
		checkbox = new Checkbox();
		add(checkbox);
		checkbox.addItemListener(this);
		active = checkbox;
		
		scrollbar = new Scrollbar(Scrollbar.HORIZONTAL, 0, 1, 0, 1024);
		scrollbar.setSize(140,20);
		scrollbar.setBlockIncrement(128);
		scrollbar.addAdjustmentListener(this);
		
		textfield = new TextField("0", 5);
		textfield.addActionListener(this);
	}
	
	public void itemStateChanged(ItemEvent e) {
		sensor.setRawValue(e.getStateChange() == ItemEvent.SELECTED ? 0 : Sensor.MAX_RAW_VALUE);
	}
	
	public void adjustmentValueChanged(AdjustmentEvent e) {
		//System.out.println(e.getValue());
		sensor.setRawValue(e.getValue());
	}
	
	
	public void actionPerformed(ActionEvent evt) {
		try {
			int v = Integer.parseInt(textfield.getText());
			sensor.setCumulativeValue(v);
		}
		catch (Exception e) {
		}
		textfield.setText(String.valueOf(sensor.getValue()));
	}
	
	
	public void pbSensorChanged(Sensor s) {
		Component c = null;
		
		switch(s.getType()) {
			case Sensor.TYPE_TOUCH:
				c = checkbox;
				break;
			case Sensor.TYPE_LIGHT:
			case Sensor.TYPE_NONE:
			case Sensor.TYPE_TEMPERATURE:
				c = scrollbar;
				break;
			case Sensor.TYPE_ROTATION:
				c = textfield;
				break;
		}
		
		if (c==active) return;
		
		if (active != null)
			remove(active);
		
		if (c!=null)
			add(c);
		
		doLayout();
		active = c;
	}
};

